; ----------------------------------------------------------------------------
; RetroForth Release 9, Revision 1
; ----------------------------------------------------------------------------

format PE console
stack 16000
entry start

include "../rx/macros"

section ".code" code readable writeable executable
start:
	call bootstrap_rx
	upsh winboot
	upsh [winboot.size]
	call eval

	jmp boot
;---------------------------------------------------------------------
old_esp  dd 0
lib_addr dd 0

code 'from', _from
        upsh 32
        call parse
        mov ecx, [esi]
        add ecx, eax
        mov byte [ecx], 0
        drop
	upop edx
	pusha
	push dword edx
	call [LoadLibrary]
	mov dword [lib_addr], eax
	popa
next

code 'import', _import
	pusha
        upsh 32
        call parse
        
        mov ecx, [esi]
        add ecx, eax
        mov byte [ecx], 0
        push dword [esi]
        call _entry
	push dword [lib_addr]
	call [GetProcAddress]
	cmp eax, 0
	jz import_error
	call literal
	popa
	call literal
	upsh _invoke
	call compile
        mov dword [imported], -1         ; success
	jmp ssemi


ixt dd 0
code 'map:', _map
	pusha
        upsh 32
        call parse
        
        mov ecx, [esi]
        add ecx, eax
        mov byte [ecx], 0
        push dword [esi]
        call _entry
	push dword [lib_addr]
	call [GetProcAddress]
	cmp eax, 0
	jz import_error
	mov dword [ixt], eax
	call literal
	popa
	upsh [ixt]
        mov dword [imported], -1         ; success
next

import_error:
        popa
        drop
        upsh error_message
        upsh [error_message.length]
        embed 'type last @ :name count type cr last @ @ last !'
        mov dword [imported], 0         ; failed
        ret
error_message db 'Error: Attempt to import unknown function: '
.length dd $-error_message

code 'imported?', import_status
        upsh [imported]
next

imported dd 0

code 'invoke', _invoke
	mov dword [old_esp], esp
	swap
	upop [faddr]		; Get the function address
	upop ecx
	cmp ecx, 0		;
	jz .run			;
.chain: push eax
	drop
	loop .chain
.run:	upsh [faddr]		; Restore the function address
	call dword eax
	mov esp, dword [old_esp]
next
faddr dd 0
;---------------------------------------------------------------------
; same as cb: but for stdcall functions:
code 'stdcb:', _stdcb
        call callback_common
        shl eax, 2  
        mov [do_cb_adjust_esp+2], al
        jmp _cb.1
; ----------------------------------------------------------------------
; cb: ( xt numparams stacksize <name> -- )
code 'cb:', _cb                                                                      
        call callback_common
        mov [do_cb_adjust_esp+2], byte 0
.1:
        drop
        ;
        mov [do_cb_xt+1], eax
        drop

	upsh 32
	call parse
        call _entry
        ;
        mov edx, do_cb
        mov ecx, do_cb$
        
.2:     mov edi, [h]
        add dword [h], ecx
.a:     mov bl, byte [edx]      ; htp123 on irc.freenode.net#fasm
        nop
        mov byte [edi], bl      ; suggested a fix for these two lines.
        inc edi                 ; It is now implemented. Thanks!
        inc edx                 ;
        loop .a                 ;
        ret

callback_common:
        shl eax, 2      ; num words --> bytes
        mov ebx, [h]
        add ebx, eax
        mov [h], ebx
        drop
        ; patch up the code;
        sub ebx, 4
        mov [do_cb_esi+1], ebx
        ; 
        mov [do_cb_numparams], eax
        ret
        
        
; Stub for a callback called from an external "C" routine.
; These values are patched up, and the result is copied into 'here'
; to be executed at run-time.
do_cb:
        pusha
do_cb_esi: ; patch ESI to our own private spot:
        mov esi, dword 0
        db $b9
do_cb_numparams: ; move parameters from ESP stack to our stack:
        dd 0
        jecxz do_cb_xt

        lea ebx, [esp+36]
do_cb_copy_params: ; loop over parameters and 'push' them
        upsh dword [ebx]
        add ebx, 4
        loop do_cb_copy_params

do_cb_xt: ; move the target XT to EBX
        mov ebx, 0
        ; and call it!
        call dword ebx
        mov [cb_eax], eax
        popa
        mov eax, [cb_eax]

do_cb_adjust_esp: ; 83 C4 00
        ret 0
        cb_eax: dd 0
do_cb$ =  $ - do_cb
         


;---------------------------------------------------------------------
winboot:
	include "../modules.inc"
	file "windows.f"
.size   dd $-winboot

hinstance rd 1
StdIn rd 1
StdOut rd 1
emit_buffer rd 1
written_buffer rd 1
tempX rd 1
;---------------------------------------------------------------------
macro library [name,string] {
  forward
    local _label
    dd RVA name#.lookup,0,0,RVA _label,RVA name#.address
  common
    dd 0,0,0,0,0
  forward
    _label db string,0
    rb RVA $ and 1
}
macro import name,[label,string] {
  common
    name#.lookup:
  forward
    local _label
    dd RVA _label
  common
    dd 0
    name#.address:
  forward
    label dd RVA _label
  common
    dd 0
  forward
    _label dw 0
    db string,0
    rb RVA $ and 1
}
;---------------------------------------------------------------------
; windows-specific linkage stuff:
data import

  library kernel,'KERNEL32.DLL'

  import kernel,\
	LoadLibrary,'LoadLibraryA',\
	GetProcAddress,'GetProcAddress'
end data

;---------------------------------------------------------------------
code 'boot', boot
next 
;---------------------------------------------------------------------
include "../rx/rx"
